#!/usr/bin/env python3

import time
import logging

from prometheus_client import start_http_server
from prometheus_client.core import REGISTRY

from technitium_exporter.collector import TechnitiumCollector
from technitium_exporter.args import parse_args

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


def main():
    args = parse_args()
    REGISTRY.register(TechnitiumCollector(args))
    logger.info(f"Starting HTTP server on {args.address}:{args.port}")
    start_http_server(args.port, args.address)
    while True:
        time.sleep(1)


if __name__ == "__main__":
    main()



PROTOCOL_TYPES = ("Tcp", "Udp", "Tls", "Https", "Quic")
COMMON_RECORD_TYPES = ("A", "AAAA", "CNAME", "MX", "NS", "PTR", "SOA", "SRV",  "TXT")

RECORD_TYPES = {
    "A": "IPv4 address",
    "AAAA": "IPv6 address",
    "AFSDB": "AFS database location",
    "APL": "Address prefix list",
    "AXFR": "Authoritative zone transfer",
    "CAA": "Certification authority authorization",
    "CDNSKEY": "Child copy of a DNSKEY",
    "CDS": "Child copy of DS",
    "CERT": "Cryptographic certificate",
    "CNAME": "Canonical name",
    "CSYNC": "Child-to-parent synchronization",
    "DHCID": "DHCP identifier",
    "DLV": "DNSSEC lookaside validation",
    "DNAME": "Delegation name",
    "DNSKEY": "Cryptographic key for DNSSEC",
    "DS": "Delegation signer",
    "EUI48": "MAC address (EUI-48)",
    "EUI64": "Mac address (EUI-64)",
    "HINFO": "Host information",
    "HIP": "Host identification protocol",
    "HTTPS": "HTTPS binding",
    "IPSECKEY": "Cryptographic key for IPsec",
    "IXFR": "Incremental zone transfer",
    "KEY": "Cryptographic key for DNSSEC (obsoleted by DNSKEY)",
    "KX": "Key exchange",
    "LOC": "Geographical location",
    "MX": "Mail exchange",
    "NAPTR": "Naming authority pointer",
    "NS": "Name server",
    "NSEC3": "Next secure (version 3)",
    "NSEC3PARAM": "Parameter for NSEC3",
    "NSEC": "Next secure (obsoleted by NSEC3)",
    "NXT": "DNSSEC key (obsoleted by NSEC)",
    "OPENPGPKEY": "Public key for OpenPGP",
    "OPT": "EDNS option",
    "PTR": "Canonical name pointer",
    "RP": "Responsible person",
    "RRSIG": "Resource record signature for DNSSEC",
    "SIG": "Resource record signature for DNSSEC (obsoleted by RRSIG)",
    "SMIMEA": "S/MIME association",
    "SOA": "Start of authority",
    "SSHFP": "Public key fingerprint for SSH",
    "SVCB": "Service binding",
    "SRV": "Service locator",
    "TA": "Trust authority for DNSSEC",
    "TKEY": "Transaction key",
    "TLSA": "Certificate association for TLS",
    "TSIG": "Transaction signature",
    "TXT": "Human-readable text",
    "URI": "Uniform resource identifier",
    "ZONEMD": "Message digest for DNS zones",
}

HELP_FOR_STATS = {
    "technitium_stats_total_queries": "Total number of DNS queries received by the server",
    "technitium_stats_total_no_error": "Total number of queries resolved without error",
    "technitium_stats_total_server_failure": "Total number of queries with SERVFAIL result",
    "technitium_stats_total_nx_domain": "Total number of queries with NXDOMAIN result",
    "technitium_stats_total_refused": "Total number of queries refused",
    "technitium_stats_total_authoritative": "Total number of queries resolved as authoritative",
    "technitium_stats_total_recursive": "Total number of queries resolved recursively",
    "technitium_stats_total_cached": "Total number of results cached",
    "technitium_stats_total_blocked": "Total number of queries blocked",
    "technitium_stats_total_dropped": "Total number of queries dropped",
    "technitium_stats_total_clients": "Number of clients using the DNS",
    "technitium_stats_zones": "Number of zones managed by the DNS",
    "technitium_stats_cached_entries": "Number of DNS entries cached by the server",
    "technitium_stats_allowed_zones": "Number of zones explicitely allowed by the DNS",
    "technitium_stats_blocked_zones": "Number of zones blocked by the DNS",
    "technitium_stats_allow_list_zones": "Number of zones in the allow lists of the DNS",
    "technitium_stats_block_list_zones": "Number of zones in the block lists of the DNS",
}


from configargparse import ArgumentParser


def parse_args():
    parser = ArgumentParser()
    parser.add_argument("--address", help="Interface to listen on", default="0.0.0.0", env_var="ADDRESS")
    parser.add_argument("--port", help="Port number to listen on", type=int, default=9080, env_var="PORT")
    parser.add_argument("--token", help="Technitium API token", required=True, env_var="TECHNITIUM_TOKEN")
    parser.add_argument("--url", help="Technitium server URL", required=True, env_var="TECHNITIUM_URL")
    parser.add_argument("--all-record-types", help="Include counts for all record types, not just the common ones",
                        action="store_true", default=False)
    return parser.parse_args()



import requests

from argparse import Namespace
from datetime import datetime, timedelta
from prometheus_client.core import GaugeMetricFamily, CounterMetricFamily
from prometheus_client.registry import Collector

from technitium_exporter.lookups import RECORD_TYPES, COMMON_RECORD_TYPES, PROTOCOL_TYPES, HELP_FOR_STATS
from technitium_exporter.constants import CAMEL_CASE


class TechnitiumCollector(Collector):
    def __init__(self, args: Namespace):
        self._args = args
        super().__init__()

    def collect_top_clients(self, stats: dict):
        fam = GaugeMetricFamily("technitium_top_clients", "Top 10 clients by number of requests",
                                labels=["rank", "name", "domain", "rateLimited"])
        for rank, top in enumerate(stats["topClients"], start=1):
            fam.add_metric([str(rank), top["name"], top.get("domain", "None"), str(top["rateLimited"])], top["hits"])
        yield fam

    def collect_top_domains(self, stats: dict):
        fam = GaugeMetricFamily("technitium_top_domains", "Top 10 domains requested for lookup",
                                labels=["rank", "name"])
        for rank, top in enumerate(stats["topDomains"], start=1):
            fam.add_metric([str(rank), top["name"]], top["hits"])
        yield fam

    def collect_top_blocked(self, stats: dict):
        fam = GaugeMetricFamily("technitium_top_blocked", "Top 10 blocked domains",
                                labels=["rank", "name"])
        for rank, top in enumerate(stats["topBlockedDomains"], start=1):
            fam.add_metric([str(rank), top["name"]], top["hits"])
        yield fam

    def collect_update_check(self):
        response = requests.get(f"{self._args.url}/api/user/checkForUpdate", params={"token": self._args.token})
        response.raise_for_status()
        json = response.json()
        assert json["status"] == "ok", "Server returned invalid status for update check"
        yield GaugeMetricFamily("technitium_update_available", "Returns 1 if there is a newer version of Technitium",
                                1 if json["response"]["updateAvailable"] else 0)

    def get_stats(self):
        now = datetime.now().isoformat()
        then = (datetime.now() - timedelta(minutes=1)).isoformat()
        response = requests.get(f"{self._args.url}/api/dashboard/stats/get",
                                params={"token": self._args.token, "type": "custom", "start": then, "end": now})
        response.raise_for_status()
        json = response.json()
        assert json["status"] == "ok", "Server returned invalid status for stats request"
        return json["response"]

    def collect_general_stats(self, stats: dict):
        for key, val in stats["stats"].items():
            name = "technitium_stats_" + CAMEL_CASE.sub("_", key).lower()
            desc = HELP_FOR_STATS[name]
            if key.startswith("total"):
                name = name.replace("_total", "")
                yield CounterMetricFamily(name, desc, val)
            else:
                yield GaugeMetricFamily(name, desc, val)

    def collect_stats_by_record_type(self, stats: dict):
        data = stats["queryTypeChartData"]
        fam = GaugeMetricFamily("technitium_record_type_count",
                                "Number of responses by record type", labels=["record_type", "desc"])
        by_type = dict()
        for record_type, count in zip(data["labels"], data["datasets"][0]["data"]):
            by_type[record_type] = count
        for record_type, help in RECORD_TYPES.items():
            if self._args.all_record_types or record_type in COMMON_RECORD_TYPES:
                fam.add_metric([record_type, help], by_type.get(record_type, 0))
        yield fam

    def collect_stats_by_protocol_type(self, stats: dict):
        data = stats["protocolTypeChartData"]
        fam = GaugeMetricFamily("technitium_protocol_type_count", "Number of requests by protocol", labels=["protocol"])
        by_proto = dict()
        for proto, count in zip(data["labels"], data["datasets"][0]["data"]):
            by_proto[proto.lower()] = count
        for proto in PROTOCOL_TYPES:
            fam.add_metric([proto], by_proto.get(proto, 0))
        yield fam

    def collect_request_stats(self, stats: dict):
        fam1 = GaugeMetricFamily("technitium_dns_request_result_count",
                                 "Number of requests with the given result", labels=["result"])
        fam2 = GaugeMetricFamily("technitium_dns_resolve_mode_count",
                                 "Number of requests resolved in a given mode", labels=["result"])
        fam3 = GaugeMetricFamily("technitium_dns_clients_connected", "Number of clients sending requests")
        for dataset in stats["mainChartData"]["datasets"]:
            name = dataset["label"].lower().replace(" ", "_")
            if name in ("total", "no_error", "server_failure", "nx_domain", "refused"):
                fam1.add_metric([name], dataset["data"][-1])
            elif name == "clients":
                fam3.add_metric([], dataset["data"][-1])
            else:
                fam2.add_metric([name], dataset["data"][-1])
        yield fam1
        yield fam2

    def collect(self):
        stats = self.get_stats()

        yield from self.collect_update_check()
        yield from self.collect_general_stats(stats)
        yield from self.collect_top_clients(stats)
        yield from self.collect_top_domains(stats)
        yield from self.collect_top_blocked(stats)
        yield from self.collect_request_stats(stats)
        yield from self.collect_stats_by_record_type(stats)
        yield from self.collect_stats_by_protocol_type(stats)
